package sach.eventmodel;

import java.util.*;
import sach.logic.Figurka.*;
import sach.logic.*;

/**
 * Trieda události posielanej v rámci komunikácie medzi komponentami
 * @author Adam Netočný
 */
public class TahEvent extends EventObject {
    /**
     * Bod odkiaľ užívateľ tiahol
     */
    private Point odkial = null;

    /**
     * Bod kam užívateľ tiahol
     */
    private Point kam = null;

    /**
     * Reprezentuje typy události
     */
    public static enum EventType { 
        /**
         * Událosť nesie informáciu o ťahu
         */
        tah,
        
        /**
         * Udalosť nesie informáciu o chybnom ťahu
         */
        chyba,

        /**
         * Udalosť nesie informáciu o výbere figúrky(pri pešiakovi na konci poľa)
         */
        vyber,
        
        /**
         * Udalosť nesie chybovú hlášku
         */
        chybovaHlaska,
        /**
         * Udalosť nesie informáciu o vzniku šachu
         */
        sach,
        /**
         * Událosť nesie informáciu o vzniku matu
         */
        mat
    }
    
    /**
     * Nesie informáciu o tom, akého typu je udalosť
     */
    private EventType typ = null;

    /**
     * Uložená informácia o farbe figúrky hráča
     */
    private Figurka.Color farba = null;

    /**
     * Uložená informácia o type figúrky hráča
     */
    private Figurka.Type typFig = null;

    /**
     * Uložená informácia o chybe(chybová hláška)
     */
    private String chyba = null;

    /**
     * Konštruktor
     * @param zdroj Objekt, na ktoróm udalosť vznikla
     * @param odkial Bod odkiaľ užívateľ tiahol
     * @param kam Bod kam užívateľ tiahol
     * @param farba Informácia o farbe figúrky hráča
     * @param typ Informácia o type udalosti
     */
    public TahEvent(Object zdroj, Point odkial, Point kam, Figurka.Color farba, EventType typ) {
        super(zdroj);
        setOdkial(odkial);
        setKam(kam);
        setTyp(typ);
        setFarba(farba);
        setChyba("");
        setTypFig(null);
    }

    /**
     * Konštruktor
     * @param zdroj Objekt, na ktoróm udalosť vznikla
     * @param odkial Bod odkiaľ užívateľ tiahol
     * @param kam Bod kam užívateľ tiahol
     * @param farba Informácia o farbe figúrky hráča
     * @param typFig Informácia o type figúrky hráča
     * @param typ Informácia o type udalosti
     */
    public TahEvent(Object zdroj, Point odkial, Point kam, Figurka.Color farba, Figurka.Type typFig, EventType typ) {
        super(zdroj);
        setOdkial(odkial);
        setKam(kam);
        setTyp(typ);
        setFarba(farba);
        setChyba("");
        setTypFig(typFig);
    }

    /**
     * Konštruktor
     * @param zdroj Objekt, na ktoróm udalosť vznikla
     * @param hlaska Chybová hláška
     * @param typ Informácia o type udalosti
     */
    public TahEvent(Object zdroj, String hlaska, EventType typ) {
        super(zdroj);
        setOdkial(null);
        setKam(null);
        setTyp(typ);
        setFarba(null);
        setChyba(hlaska);
        setTypFig(null);
    }

    /**
     * Konštruktor
     * @param zdroj Objekt, na ktoróm udalosť
     * @param farba Informácia o farbe figúrky hráča
     * @param typ Informácia o type udalosti
     */
    public TahEvent(Object zdroj, Figurka.Color farba, EventType typ) {
        super(zdroj);
        setOdkial(null);
        setKam(null);
        setTyp(typ);
        setFarba(farba);
        setChyba("");
        setTypFig(null);
    }

    /**
     * Vráti bod kam užívateľ tiahol
     * @return Bod kam užívateľ tiahol
     */
    public Point getKam() {
        return kam;
    }

    /**
     * Vráti bod odkiaľ užívateľ tiahol
     * @return Bod odkiaľ užívateľ tiahol
     */
    public Point getOdkial() {
        return odkial;
    }

    /**
     * Vráti informáciu o tom, akého typu je udalosť
     * @return Informácia o tom, akého typu je udalosť
     */
    public EventType getTyp() {
        return typ;
    }

    /**
     * Vráti informáciu o farbe figúrky hráča
     * @return Informácia o farbe figúrky hráča
     */
    public Color getFarba() {
        return farba;
    }

    /**
     * Vráti informáciu o type figúrky hráča
     * @return Informácia o type figúrky hráča
     */
    public Type getTypFig() {
        return typFig;
    }

    /**
     * Vráti chybovú hlášku
     * @return Chybová hláška
     */
    public String getChyba() {
        return chyba;
    }



    /**
     * Nastaví bod kam užívateľ tiahol
     * @param kam Bod kam užívateľ tiahol
     * @deprecated Nie je potrebná - lepšie je nahradiť vytvorením nového objektu
     */
    @Deprecated
    public void setKam(Point kam) {
        this.kam = kam;
    }

    /**
     * Nastaví bod odkiaľ užívateľ tiahol
     * @param odkial Bod odkiaľ užívateľ tiahol
     * @deprecated Nie je potrebná - lepšie je nahradiť vytvorením nového objektu
     */
    @Deprecated
    public void setOdkial(Point odkial) {
        this.odkial = odkial;
    }

    /**
     * Nastaví informáciu o tom, akého typu je udalosť
     * @param typ Informácia o tom, akého typu je udalosť
     * @deprecated Nie je potrebná - lepšie je nahradiť vytvorením nového objektu
     */
    @Deprecated
    public void setTyp(EventType typ) {
        this.typ = typ;
    }

    /**
     * Nastaví informáciu o farbe figúrky hráča
     * @param farba Informácia o farbe figúrky hráča
     * @deprecated Nie je potrebná - lepšie je nahradiť vytvorením nového objektu
     */
    @Deprecated
    public void setFarba(Color farba) {
        this.farba = farba;
    }

    /**
     * Nastaví informáciu o type figúrky hráča
     * @param typFig Informácia o type figúrky hráča
     * @deprecated Nie je potrebná - lepšie je nahradiť vytvorením nového objektu
     */
    @Deprecated
    public void setTypFig(Type typFig) {
        this.typFig = typFig;
    }

    /**
     * Nastaví chybovú hlášku
     * @param chyba Chybová hláška
     * @deprecated Nie je potrebná - lepšie je nahradiť vytvorením nového objektu
     */
    @Deprecated
    public void setChyba(String chyba) {
        this.chyba = chyba;
    }

    /**
     * Porovnanie dvoch objektov pomocou hashCode
     * @param obj Objekt, s ktorým chceme porovnávať
     * @return true - zhodné; false - rozdielne
     */
    @Override
    public boolean equals(Object obj) {
        if(obj != null){
            try {
                if (obj.getClass() == this.getClass()) {
                    return ((TahEvent)obj).hashCode() == hashCode();
                }
            } catch (Exception ex) {
                return false;
            }
        }
        return false;
    }

    /**
     * Prekrytie metódy hashCode - pridáva rozlíšenie podľa typu události
     * @return "hash"
     */
    @Override
    public int hashCode() {
        int hash = 7;
        if(typ == EventType.tah || typ == EventType.chyba) {
            hash = 17 * hash + (this.odkial != null ? this.odkial.hashCode() : 0);
            hash = 17 * hash + (this.kam != null ? this.kam.hashCode() : 0);
        }
        if(typ == EventType.sach || typ == EventType.mat) {
            hash = 17 * hash + (this.farba != null ? this.farba.hashCode() : 0);
        }
        if(typ == EventType.vyber){
            hash = 17 * hash + (this.odkial != null ? this.odkial.hashCode() : 0);
            hash = 17 * hash + (this.kam != null ? this.kam.hashCode() : 0);
            hash = 17 * hash + (this.farba != null ? this.farba.hashCode() : 0);
        }
        if(typ == EventType.chybovaHlaska) {
            hash = 17 * hash + (this.chyba != null ? this.chyba.hashCode() : 0);
        }
        return hash;
    }
}
