package sach.logic;

/**
 * Spoločný predok všetkých figúriek, implementuje spoločné vlastnosti
 * @author Adam Netočný
 */
public abstract class Figurka implements FigurkaInterface {
    /**
     * X-ová pozícia
     */
    protected int x;
    /**
     * Y-ová pozícia
     */
    protected int y;

    /**
     * Reprezentuje farby figúriek
     */
    public static enum Color {

        /**
         * Figúrka má bielu farbu
         */
        biela,
        /**
         * Figúrka má čiernu farbu
         */
        cierna
    };

    /**
     * Reprezentuje typy figúriek
     */
    public static enum Type {
        /**
         * Figúrka je typu pešiak
         */
        pesiak,
        /**
         * Figúrka je typu strelec
         */
        strelec,
        /**
         * Figúrka je typu kôň
         */
        kon,
        /**
         * Figúrka je typu veža
         */
        veza,
        /**
         * Figúrka je typu dáma
         */
        dama,
        /**
         * Figúrka je typu kráľ
         */
        kral
    }

    /**
     * Farba danej figúrky
     */
    protected Color farba;
    /**
     * Typ danej figúrky
     */
    protected Type typ;

    /**
     * Pešiak došiel na koniec poľa a musí byť nahradený za inú figúrku
     */
    public class PesiakNaKonciPola extends Throwable {
        /**
         * Figúrka, na ktorej vznikla daná vynínmka
         */
        public FigurkaInterface figurka;

        /**
         * Vytvorí novú vynínmku podľa potrebných údajov
         * @param figurka Figúrka, na ktorej vznikla daná vynínmka
         */
        PesiakNaKonciPola(FigurkaInterface figurka) {
            this.figurka = figurka;
        }
    }

    /**
     * Rozšáda - kráľ žiada rozšádu s vežou
     */
    public class Rozsada extends Throwable {
        /**
         * Figúrka, na ktorej vznikla daná vynínmka
         */
        public FigurkaInterface figurka;
        /**
         * Vytvorí novú vynínmku podľa potrebných údajov
         * @param figurka Figúrka, na ktorej vznikla daná vynínmka
         */
        Rozsada(FigurkaInterface figurka) {
            this.figurka = figurka;
        }
    }

    /**
     * Všeobecný konštuktor - nastavuje potrebné údaje o figúrke
     * @param x X-ová pozícia, na ktorej je figúrka umiestnená
     * @param y Y-ová pozícia, na ktorej je figúrka umiestnená
     * @param farba Farba novovytvorenej figúrky
     */
    Figurka(int x, int y, Color farba){
        this.x = x;
        this.y = y;
        this.farba = farba;
    }

    /**
     * Kontrola ťahu figúrkou
     * @param x X-ová pozícia, na ktorú by sa mala figúrka presunúť
     * @param y Y-ová pozícia, na ktorú by sa mala figúrka presunúť
     * @param sach Udáva pole, v ktorom sú rozmiestnené figúrky pričom sa berie v úvahu aj rozmiestnenie ostatných figúriek
     * @return true, ak sa môže figúrka presunúť, inak vráti false
     * @throws PesiakNaKonciPola
     * @throws Rozsada
     */
    public boolean tah(int x, int y, FigurkaInterface[][] sach) throws PesiakNaKonciPola, Rozsada {
        return tah(x, y, sach, true);
    }

    /**
     * Kontrola ťahu figúrkou
     * @param x X-ová pozícia, na ktorú by sa mala figúrka presunúť
     * @param y Y-ová pozícia, na ktorú by sa mala figúrka presunúť
     * @param sach Udáva pole, v ktorom sú rozmiestnené figúrky pričom sa berie v úvahu aj rozmiestnenie ostatných figúriek
     * @param detKral Určuje či má, alebo nemá brať v úvahu, že figúrkou, na ktorú sa snaží preisť je kráľ
     * @return true, ak sa môže figúrka presunúť, inak vráti false
     * @throws PesiakNaKonciPola
     * @throws Rozsada
     */
    public abstract boolean tah(int x, int y, FigurkaInterface[][] sach, boolean detKral) throws PesiakNaKonciPola, Rozsada;

    /**
     * Vráti x-ovú pozíciu figúrky v rámci poľa
     * @return X-ová pozícia figúrky v rámci poľa
     */
    public int getX(){
        return this.x;
    }

    /**
     * Vráti y-ovú pozíciu figúrky v rámci poľa
     * @return Y-ová pozícia figúrky v rámci poľa
     */
    public int getY(){
        return this.y;
    }

    /**
     * Vráti farbu figúrky
     * @return Farba figúrky
     */
    public Color getColor(){
        return this.farba;
    }

    /**
     * Vráti typ figúrky
     * @return Typ figúrky
     */
    public Type getType(){
        return this.typ;
    }

    /**
     * Oznámi figúrke, na akej pozícii sa nachádza
     * @param x X-ová pozícia figúrky v rámci poľa
     * @param y Y-ová pozícia figúrky v rámci poľa
     */
    public void setXY(int x, int y){
        this.x = x;
        this.y = y;
    }

    /**
     * Hľadá najbližšiu figúrku v X-ovom smere podľa určenia smeru
     * @param sach Udáva pole, v ktorom sú rozmiestnené figúrky pričom sa berie v úvahu aj rozmiestnenie ostatných figúriek
     * @param smer true - v smere dolu, false - v smere hore
     * @return Najbližšia figúrka v X-ovom smere, ak existuje, inak null
     */
    protected FigurkaInterface getFigrkaDirectionX(FigurkaInterface[][] sach, boolean smer){
        boolean nasiel = false;

        int smr = 0;
        if(smer){
            smr = 1;
        } else {
            smr = -1;
        }

        int a = this.x+smr;

        for(; a<8 && a>=0; a+=smr){
            if(sach[a][this.y]!=null){
                nasiel=true;
                break;
            }
        }
        if(nasiel){
            return sach[a][this.y];
        } else {
            return null;
        }
    }

    /**
     * Hľadá najbližšiu figúrku v Y-ovom smere podľa určenia smeru
     * @param sach Udáva pole, v ktorom sú rozmiestnené figúrky pričom sa berie v úvahu aj rozmiestnenie ostatných figúriek
     * @param smer true - v smere doprava, false - v smere doľava
     * @return Najbližšia figúrka v Y-ovom smere, ak existuje, inak null
     */
    protected FigurkaInterface getFigrkaDirectionY(FigurkaInterface[][] sach, boolean smer){
        boolean nasiel = false;

        int smr = 0;
        if(smer){
            smr = 1;
        } else {
            smr = -1;
        }
        
        int a = this.y+smr;

        for(; a<8 && a>=0; a+=smr){
            if(sach[this.x][a]!=null){
                nasiel=true;
                break;
            }
        }
        if(nasiel){
            return sach[this.x][a];
        } else {
            return null;
        }
    }

    /**
     * Hľadá najbližšiu figúrku v diagonálnom smere podľa určenia smeru
     * @param sach Udáva pole, v ktorom sú rozmiestnené figúrky pričom sa berie v úvahu aj rozmiestnenie ostatných figúriek
     * @param doprava true - v smere doprava, false - v smere doľava
     * @param dole true - v smere dole, false - v smere hore
     * @return Najbližšia figúrka v diagonálnom smere, ak existuje, inak null
     */
    protected FigurkaInterface getFigrkaDirectionXY(FigurkaInterface[][] sach, boolean doprava, boolean dole){
        boolean nasiel = false;

        int xplus = 0;
        int yplus = 0;

        if(doprava && dole){
            xplus = 1;
            yplus = 1;
        } else if(doprava && !dole){
            xplus = -1;
            yplus = 1;
        } else if(!doprava && dole){
            xplus = 1;
            yplus = -1;
        } else if(!doprava && !dole){
            xplus = -1;
            yplus = -1;
        }
        
        int a = this.y+yplus;
        int b = this.x+xplus;
        while((a<=7 && a>=0) && (b<=7 && b>=0)){
            if(sach[b][a]!=null){
                nasiel=true;
                break;
            }
            a+=yplus;
            b+=xplus;
        }
        if(nasiel){
            return sach[b][a];
        } else {
            return null;
        }
    }
}
