package sach.network;

import java.io.*;
import java.net.*;
import java.util.ArrayList;
import sach.SachApp;
import sach.eventmodel.*;
import sach.logic.Figurka.*;
import sach.logic.*;

/**
 * Implementuje rozhrannie sieťovej hry
 * @author Adam Netočný
 */
public abstract class NetworkGame extends TahHandler implements NetworkGameInterface {
    /**
     * Aktuálne pripojenie
     */
    protected Socket connection;

    /**
     * Zoznam vsetkych vlaken spustenych z daneho komponentu
     */
    private ArrayList<Runnable> l = new ArrayList<Runnable>();
    /**
     * OutputStream pripojenia
     */
    protected OutputStream outStream;

    /**
     * InputStream pripojenia
     */
    protected InputStream inStream;

    /**
     * Obsahuje aktuálne odosielanú alebo príjmanú správu
     */
    protected String message;

    /**
     * Obsahuje názov figúrky, ak nejaký prijal; inak prazdny reťazec
     */
    protected String figMessage = "";

    /**
     * Host Name servru
     */
    protected String hostName;
    /**
     * IP adresa servru
     */
    protected String ipAddress;

    /**
     * Časový limit pripojenia
     */
    private static int TIME_OUT = 600000;
    
    /**
     * Prenáša informáciu o bodoch medzi vláknami
     */
    private static Point[] body = new Point[2];

    /**
     * Základná výnimka sieťovej hry
     */
    public class NetworkGameException extends Throwable {
        /**
         * Obsahuje internú správu výnimky
         */
        private String internalMessage;
        /**
         * Vytvorí výnimku so štandardným textom
         */
        public NetworkGameException() {
            internalMessage = "Vyskytla sa necakana chyba v sietovej hre!";
        }
        /**
         * Vytvorí výnimku so špecifickým textom
         * @param message Správa, ktorú má odoslať
         */
        public NetworkGameException(String message) {
            internalMessage = "Vyskytla sa necakana chyba v sietovej hre: " + message;
        }

        /**
         * Vracia internú správu výnimky
         * @return Interná správa výnimky
         */
        @Override
        public String getMessage(){
            return internalMessage;
        }
    }

    /**
     * Výnimka pre vypršanie časového limitu
     */
    public class NetworkTimeOutException extends Throwable {
        /**
         * Obsahuje internú správu výnimky
         */
        private String internalMessage;
        /**
         * Vytvorí výnimku so štandardným textom
         */
        public NetworkTimeOutException() {
            internalMessage = "Casovy limit spojenia vyprsal!";
        }

        /**
         * Vracia internú správu výnimky
         * @return Interná správa výnimky
         */
        @Override
        public String getMessage(){
            return internalMessage;
        }
    }

    /**
     * Výnimka pre prerušenie pripojenia
     */
    public class NetworkConnectionRefusedException extends Throwable {
        /**
         * Obsahuje internú správu výnimky
         */
        private String internalMessage;
        /**
         * Vytvorí výnimku so štandardným textom
         */
        public NetworkConnectionRefusedException() {
            internalMessage = "Spojenie bolo neocakavane prerusene!";
        }

        /**
         * Vracia internú správu výnimky
         * @return Interná správa výnimky
         */
        @Override
        public String getMessage(){
            return internalMessage;
        }
    }

    /**
     * Vytvorí pripojenie pre sieťovú hru
     * @param ip IP adresa, ne ktorú sa má pripojiť
     * @param port Číslo portu
     * @return true - úspešne pripojený, false - nepripojený
     * @throws NetworkGameException
     * @throws NetworkTimeOutException
     * @throws NetworkConnectionRefusedException
     */
    public abstract boolean connect(String ip, int port) throws NetworkGameException, NetworkTimeOutException, NetworkConnectionRefusedException;

    /**
     * Kontroluje, či je hra pripojená
     * @return true - pripojená, false - odpojená
     */
    public abstract boolean isConnected();

    /**
     * Ukončí pripojenie
     * @throws NetworkGameException
     */
    protected abstract void closeConnection() throws NetworkGameException;

    /**
     * Vráti aktuálnu IP adresu serveru
     * @return IP adresa servru
     */
    public String getIpAddress() {
        return ipAddress;
    }

    /**
     * Vráti Host Name servru
     * @return Host Name servru
     */
    public String getHostName() {
        return hostName;
    }

    /**
     * Odošle dvojicu reprezentujúcu pozíciu na hracom poli
     * @param x X-ová súradnica
     * @param y Y-ová súradnica
     * @throws sach.NetworkGame.NetworkGameException
     * @throws sach.NetworkGame.NetworkTimeOutException
     * @throws sach.NetworkGame.NetworkConnectionRefusedException
     */
    private void sendTah(int x, int y) throws NetworkGameException, NetworkTimeOutException, NetworkConnectionRefusedException  {
        if(!isConnected()){
            throw new NetworkGameException("Program nie je pripojeny!");
        }

        write(x+":"+y);
        
        if(!getResponse()){
            throw new NetworkGameException("Program nedostal odozvu po odoslani dat!");
        }
    }

    /**
     * Príjme dvojicu reprezentujúcu pozíciu na hracom poli
     * @return Bod určujúci danu pozíciu
     * @throws sach.NetworkGame.NetworkGameException
     * @throws sach.NetworkGame.NetworkTimeOutException
     * @throws sach.NetworkGame.NetworkConnectionRefusedException
     */
    private Point reciveTah() throws NetworkGameException, NetworkTimeOutException, NetworkConnectionRefusedException {
        if(!isConnected()){
            throw new NetworkGameException("Program nie je pripojeny!");
        }

        String[] xy;
        read();
        if(message.contains(":")){
            xy = message.split(":");
            sendResponse();
            return new Point(Integer.parseInt(xy[0]), Integer.parseInt(xy[1]));
        }
        throw new NetworkGameException("Program nedostal pozadovane data!");
    }

    /**
     * Číta odozvu z pripojenia
     * @throws NetworkGameException 
     * @throws NetworkTimeOutException
     * @throws NetworkConnectionRefusedException
     */
    protected void read() throws NetworkGameException, NetworkTimeOutException, NetworkConnectionRefusedException {
        message = "";
        String l_strOutput = "";
        String l_strTemp;
        int l_iRetval = 0;
        int timeout = TIME_OUT;

        byte[] receiveBuffer = new byte[512 * 2];

        l_strOutput = "";
        try {
            //caka na data
            while(inStream.available() == 0 && isConnected()){
                Thread.sleep(200);
                timeout -= 200;
                if(timeout <= 0){
                    throw new NetworkTimeOutException();
                }
            }
            
        } catch (Exception ex) {
            throw new NetworkGameException();
        }
        try {
            //prijma data kym su posielane
            for (; (l_iRetval = inStream.read(receiveBuffer)) > 0;) {
                byte[] tmp = new byte[l_iRetval];
                for (int axb = 0; axb < l_iRetval; axb++) {
                    tmp[axb] = receiveBuffer[axb];
                }
                l_strTemp = new String(tmp, "ASCII");
                l_strOutput += l_strTemp.toString();
                if (inStream.available() == 0) {
                    break;
                }
            }
        } catch (IOException ex) {
            throw new NetworkGameException("Chyba pri citani dat!");
        }


        //zisti viacnasoby zapis a zapise hodnotu pre message
        String[] msg = l_strOutput.replaceAll("\r", "").split("\n");
        if(!msg[0].equals("")){
            this.message = msg[0];
        } else {
            throw new NetworkGameException("Program neobdrzal data!");
        }
        
        ArrayList<String> msgWithoutTestAndFigurkaMessages = new ArrayList<String>();
        for(int a = 0; a < msg.length; a++){
            if(!msg[a].equals("TEST")){
                if(msg[a].equals("KON")){
                    figMessage = msg[a];
                } else if(msg[a].equals("STRELEC")){
                    figMessage = msg[a];
                } else if(msg[a].equals("VEZA")){
                    figMessage = msg[a];
                } else if(msg[a].equals("DAMA")){
                    figMessage = msg[a];
                } else if(msg[a].equals("QUIT")){
                    closeConnection();
                    throw new NetworkConnectionRefusedException();
                } else {
                    msgWithoutTestAndFigurkaMessages.add(msg[a]);
                }                
            }
        }

        if(msg.length != msgWithoutTestAndFigurkaMessages.size()){
            msg = new String[msgWithoutTestAndFigurkaMessages.size()];
            for(int a = 0; a < msgWithoutTestAndFigurkaMessages.size(); a++){
                msg[a] = msgWithoutTestAndFigurkaMessages.get(a);
            }
            if(msg.length!=0){
                if(!msg[0].equals("")){
                    this.message = msg[0];
                } else {
                    read();
                }
            } else {
                read();
            }
        }

    if(msg.length>2 || (msg.length==2 && !msg[1].equals(""))){
            throw new NetworkGameException("Precitala sa dvojita odozva!");
        }
    }

    /**
     * Zapisuje dáta na pripojenie
     * @param messg Správa, ktorú má odoslať
     * @return true - Úspešne odoslané, inak false
     * @throws NetworkGameException
     */
    protected boolean write(String messg) throws NetworkGameException{
        message = new String(messg);
        try {
            byte[] cmdBytes = ((messg.trim() + "\r\n")).getBytes();
            outStream.write(cmdBytes, 0, cmdBytes.length);
            return true;
        } catch (Exception ex) {
            throw new NetworkGameException("Chyba pri zapise dat!");
        }
    }

    /**
     * Odošlé odozvu na pripojenie
     * @return true - Úspešne odoslané, inak false
     * @throws NetworkGameException
     */
    protected boolean sendResponse() throws NetworkGameException{
        try {
            boolean b = write("Done");
            Thread.sleep(1000);
            return b;
        } catch (InterruptedException ex) {
            return false;
        }
    }

    /**
     * Čaká na odozvu z pripojenia
     * @return true - Prijal správnu odozvu, inak false
     * @throws NetworkGameException
     * @throws NetworkTimeOutException
     * @throws NetworkConnectionRefusedException
     */
    protected boolean getResponse() throws NetworkGameException, NetworkTimeOutException, NetworkConnectionRefusedException {
        read();
        if(message.equals("Done")){
            return true;
        }
        return false;
    }

    /**
     * Uzavrie spojenie a zruší objekt
     * @throws java.lang.Throwable
     */
    @Override
    protected void finalize() throws Throwable {
        closeConnection();
        super.finalize();
    }

    /**
     * Vracia typ figúrky, ktorý zvolil užívatel na druhej strane
     * @return Typ figúrky ak nejaký došiel, inak null
     */
    private Type getFigurka(){
        if(figMessage.equals("")){
            return null;
        } else {
            if(figMessage.equals("KON")){
                figMessage = "";
                return Type.kon;
            } else if(figMessage.equals("STRELEC")){
                figMessage = "";
                return Type.strelec;
            } else if(figMessage.equals("VEZA")){
                figMessage = "";
                return Type.veza;
            } else if(figMessage.equals("DAMA")){
                figMessage = "";
                return Type.dama;
            }
            figMessage = "";
            return null;
        }
    }

    /**
     * Posiela typ figúrky, ktorý zvolil užívatel, druhej strane
     * @param fig Typ figúrky, ktorý zvolil užívaťel
     * @throws NetworkGame.NetworkGameException
     * @throws NetworkGame.NetworkTimeOutException
     * @throws NetworkGame.NetworkConnectionRefusedException 
     */
    private void setFigurka(Type fig) throws NetworkGameException, NetworkTimeOutException, NetworkConnectionRefusedException {
        if(!isConnected()){
            throw new NetworkGameException("Program nie je pripojeny!");
        }

        if(fig == Type.kon){
            write("KON");
        } else if(fig == Type.strelec){
            write("STRELEC");
        } else if(fig == Type.veza){
            write("VEZA");
        } else if(fig == Type.dama){
            write("DAMA");
        }
        
    }

    /**
     * Volaná pri prijatí události typu ťah
     * @param event Daná udalosť
     */
    @Override
    public void tah(TahEvent event) {
        if(!event.equals(this.event)){
            body[0] = event.getOdkial();
            body[1] = event.getKam();
            Vlakno vl = new Vlakno(true, true, this);
            Thread thr = new Thread(vl);
            thr.start();
            fireTah(event.getOdkial(), event.getKam(), event.getFarba());
        }
    }

    /**
     * Volaná pri prijatí události typu chybný ťah
     * @param event Daná udalosť
     */
    @Override
    public void chybyTah(TahEvent event) {
        if(!event.equals(this.event)){
            fireChybnyTah(event.getOdkial(), event.getKam(), event.getFarba());
        }
    }

    /**
     * Volaná pri prijatí události typu výber figúrky
     * @param event Daná udalosť
     */
    @Override
    public void vyberFigurku(TahEvent event) {
        if(!event.equals(this.event)){
            try {
                if(event.getTypFig() != null){
                    setFigurka(event.getTypFig());
                    body[0] = event.getOdkial();
                    body[1] = event.getKam();
                    Vlakno vl = new Vlakno(true, true, this);
                    Thread thr = new Thread(vl);
                    thr.start();
                    fireVyberFigurku(event.getOdkial(), event.getKam(), event.getTypFig(), event.getFarba());
                } else {
                    fireVyberFigurku(event.getOdkial(), event.getKam(), getFigurka(), event.getFarba());
                }
            } catch (NetworkGameException ex) {
                fireChyba(ex.getMessage() +
                        "\n\nHra sa okmzite ukonci!");
            } catch (NetworkTimeOutException ex) {
                fireChyba(ex.getMessage() +
                        "\n\nHra sa okmzite ukonci!");
            } catch (NetworkConnectionRefusedException ex) {
                fireChyba(ex.getMessage() +
                        "\n\nHra sa okmzite ukonci!");
            }
        }
    }

    /**
     * Volaná pri prijatí události typu chybová hláška
     * @param event Daná udalosť
     */
    @Override
    public void chyba(TahEvent event) {
        if(!event.equals(this.event)){
            fireChyba(event.getChyba());
        }
    }

    /**
     * Volaná pri prijatí události typu šach
     * @param event Daná udalosť
     */
    @Override
    public void sach(TahEvent event) {
        if(!event.equals(this.event)){
            fireSach(event.getFarba());
        }
    }

    /**
     * Volaná pri prijatí události typu mat
     * @param event Daná udalosť
     */
    @Override
    public void mat(TahEvent event) {
        if(!event.equals(this.event)){
            fireMat(event.getFarba());
        }
    }

    /**
     * Volaná na komponente pri ukončovaní programu
     */
    public void exit(){
        try {
            closeConnection();
        } catch (NetworkGameException ex) {
        }
    }

    /**
     * Spustí vlákno, ktoré čaká na prijatie prvého ťahu
     */
    protected void waitForFirst(){
        Vlakno vl = new Vlakno(true, false, this);
        Thread thr = new Thread(vl);
        thr.start();
    }

    //<editor-fold defaultstate="collapsed" desc="Praca cez sietove rozhrannie vo vlaknach">
    /**
     * Vlákno pre odosielanie a príjmanie dát
     */
    private class Vlakno implements Runnable {
        /**
         * Logická hodnota, či má hra očakávať dáta
         */
        private boolean recive;

        /**
         * Logická hodnota, či má hra odosielať dáta
         */
        private boolean send;

        /**
         * Odkaz na sieťové rozhrannie
         */
        private NetworkGame parent;

        /**
         * Vytvorí nové vlákno pre odosielanie a príjmanie dát
         * @param recive Logická hodnota, či má hra očakávať dáta
         * @param send Logická hodnota, či má hra odosielať dáta
         * @param parent Odkaz na sieťové rozhrannie
         */
        public Vlakno(boolean recive, boolean send, NetworkGame parent) {
            this.recive = recive;
            this.send = send;
            this.parent = parent;
        }

        /**
         * Odošle a píjme dáta
         */
        @Override
        public void run() {
            l.add(this);
            if(send){
                send(body[0].getX(), body[0].getY(), body[1].getX(), body[1].getY());
            }

            Point a = null;
            Point b = null;
            if(recive){
                a = recive();
                b = recive();
            
                body[0]=a;
                body[1]=b;
                if(a != null && b != null) {
                    fireTah(body[0], body[1], SachApp.gameStatus.getOponentColor());
                } else {
                    fireChyba("Hra nedostala pozadovane udaje!");
                }
            }
            l.remove(this);
        }
        
        //<editor-fold defaultstate="collapsed" desc="Praca cez sietove rozhrannie s osetrenim vynimiek">
        /**
         * Čaká, pokým nedostane dáta; používa sieťové rozhrannie a zároveň ošetruje vynímky
         * @return Vracia obdržaný bod na šachovnici
         */
        private Point recive(){
            try {
                return reciveTah();
            } catch (NetworkGameException ex) {
                send = false;
                recive = false;
                fireChyba(ex.getMessage() +
                        "\n\nHra sa okmzite ukonci!");
                return null;
            } catch (NetworkTimeOutException ex) {
                send = false;
                recive = false;
                fireChyba(ex.getMessage() +
                        "\n\nHra sa okmzite ukonci!");
                return null;
            } catch (NetworkConnectionRefusedException ex) {
                send = false;
                recive = false;
                fireChyba(ex.getMessage() +
                        "\n\nHra sa okmzite ukonci!");
                return null;
            }
        }

        /**
         * Odosiela dáta; používa sieťové rozhrannie a zároveň ošetruje vynímky
         * @param posX X-ová pozícia figúrky, ktorou chceme tiahnuť
         * @param posY y-ová pozícia figúrky, ktorou chceme tiahnuť
         * @param x X-ová pozícia miesta, na ktoré chceme tiahnuť
         * @param y Y-ová pozícia miesta, na ktoré chceme tiahnuť
         */
        private void send(int posX, int posY, int x, int y){
            try {
                sendTah(posX, posY);
                sendTah(x, y);
            } catch (NetworkGameException ex) {
                fireChyba(ex.getMessage() +
                        "\n\nHra sa okmzite ukonci!");
            } catch (NetworkTimeOutException ex) {
                fireChyba(ex.getMessage() +
                        "\n\nHra sa okmzite ukonci!");
            } catch (NetworkConnectionRefusedException ex){
                fireChyba(ex.getMessage() +
                        "\n\nHra sa okmzite ukonci!");
            }
        }
        //</editor-fold>
    }
    //</editor-fold>
}
